import pytest
import numpy as np
import changepoynt.algorithms.sst as ssts
import logging


class TestSST:
    def setup_method(self):
        # set a random seed
        np.random.seed(3455)

        # create a random steps signal of a certain length
        self.signal_length = 300
        x0 = 1 * np.ones(self.signal_length) + np.random.rand(self.signal_length) * 1
        x1 = 3 * np.ones(self.signal_length) + np.random.rand(self.signal_length) * 2
        x2 = 5 * np.ones(self.signal_length) + np.random.rand(self.signal_length) * 1.5
        x = np.hstack([x0, x1, x2])
        x += np.random.rand(x.size)
        self.signal = x

    def teardown_method(self):
        pass

    def test_all_methods(self):
        # initialize random default method
        sst = ssts.SST(30)

        # get the different method names
        methods = list(sst.methods.keys())

        # go through the methods and check execution
        for method in methods:
            ssts.SST(50, rank=2, method=method).transform(self.signal)

    def test_svd_method(self):

        # initialize the scorer
        sst = ssts.SST(50, method="svd")
        sst.transform(self.signal)

    def test_fbrsvd_method(self):

        # initialize the scorer
        sst = ssts.SST(50, method="fbrsvd")
        sst.transform(self.signal)

    def test_rsvd_method(self):

        # initialize the scorer
        sst = ssts.SST(50, method="rsvd")
        res = sst.transform(self.signal)
        sst = ssts.SST(50, method="rsvd", use_fast_hankel=True)
        res2 = sst.transform(self.signal)

    def test_ika_method(self):

        # initialize the scorer
        sst = ssts.SST(50, method="ika")
        res = sst.transform(self.signal)
        sst = ssts.SST(50, method="ika", use_fast_hankel=True)
        res2 = sst.transform(self.signal)

    def test_default(self):
        ssts.SST(min(5, self.signal_length//2), rank=2).transform(self.signal)

    def test_unknown_method(self):
        with pytest.raises(ValueError):
            ssts.SST(10, method='asdafwegrhqh')


if __name__ == "__main__":
    pytest.main()



# this file contains utility functions for handling and computing matrices needed for some of the algorithms in this
# package
import numpy as np
import numba as nb

import scipy as sp
import scipy.linalg as splg
import scipy.sparse.linalg
import fbpca


# @nb.jit(nopython=True)
def power_method(a_matrix: np.ndarray, x_vector: np.ndarray, n_iterations: int) -> (float, np.ndarray):
    """
    This function searches the largest (dominant) eigenvalue and corresponding eigenvector by repeated multiplication
    of the matrix A with an initial vector. It assumes a dominant eigenvalue bigger than the second one, otherwise
    it won't converge.

    For proof and explanation look at:
    https://pythonnumericalmethods.berkeley.edu/notebooks/chapter15.02-The-Power-Method.html
    :param a_matrix: 2D-Matrix of size NxN filled with floats
    :param x_vector: Vector of size Nx1 filled with floats
    :param n_iterations: the amount of iterations for the approximation
    :return: the dominant eigenvalue and corresponding eigenvector
    """

    # go through the iterations and continue to scale the returned vector, so we do not reach extreme values
    # during the iteration we scale the vector by its maximum as we can than easily extract the eigenvalue
    # TODO: This only works for our symmetric correlation matrices but not for any arbitrary matrices
    for _ in range(n_iterations):
        # multiplication with a_matrix.T @ a_matrix as can be seen in explanation of
        # https://numpy.org/doc/stable/reference/generated/numpy.linalg.svd.html
        x_vector = a_matrix @ x_vector

        # scale the vector so we keep the values in bound
        x_vector = x_vector / np.max(x_vector)

    # get the normed eigenvector
    x_vector = x_vector / np.linalg.norm(x_vector)

    # get the corresponding eigenvalue
    eigenvalue = np.linalg.norm(a_matrix @ x_vector)
    return eigenvalue, (a_matrix @ x_vector) / eigenvalue


# @jit(nopython=True)
def lanczos(a_matrix: np.ndarray, r_0: np.ndarray, k: int) -> (np.ndarray, np.ndarray):
    """
    This function computes the tri-diagonalization matrix from the square matrix C which is the result of the lanczos
    algorithm.

    The algorithm has been described and proven in:
    Idé, Tsuyoshi, and Koji Tsuda.
    "Change-point detection using krylov subspace learning."
    Proceedings of the 2007 SIAM International Conference on Data Mining.
    Society for Industrial and Applied Mathematics, 2007.

    :param a_matrix: 2D-Matrix of size NxN filled with floats where we want to find the krylov subspace approx. for
    :param r_0: intial starting vector for the subspace approximation
    :param k: size of the approximation
    :return: Returns the alpha and beta values for the tridiagonal, symmetric matrix T. alphas are the values from the
    main diagonal and beta from the off diagonal as described in
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.linalg.eigh_tridiagonal.html (alpha = d, beta = e)
    """

    # save the initial vector
    r_i = r_0
    q_i = np.zeros_like(r_i)

    # initialization of the diagonal elements
    alphas = np.zeros(shape=(k + 1,), dtype=np.float64)
    betas = np.ones(shape=(k + 1,), dtype=np.float64)

    # Subroutine 1 of the paper
    for j in range(k):
        # compute r_(j+1)
        new_q = r_i / betas[j]

        # compute the new alpha
        alphas[j + 1] = new_q.T @ (a_matrix @ new_q)

        # compute the new r
        r_i = a_matrix @ new_q - alphas[j + 1] * new_q - betas[j] * q_i

        # compute the next beta
        betas[j + 1] = np.linalg.norm(r_i)

        # update the previous q
        q_i = new_q

    return alphas[1:], betas[1:-1]


def tridiagonal_eigenvalues(alphas: np.ndarray, betas: np.ndarray, amount=-1):
    """
    This function uses a fast approach for symmetric tridiagonal matrices to calculate the [amount] highest eigenvalues
    and corresponding eigenvectors.

    :param alphas: main diagonal elements
    :param betas: off diagonal elements
    :param amount: The number of eigenvalues you want to compute (from the highest)
    :return: eigenvalues and corresponding eigenvectors
    """

    # check whether we need to use default parameters
    if amount < 0:
        amount = alphas.shape[0]

    # assertions about shape and dimensions as well as number of eigenvectors
    assert 0 < amount <= alphas.shape[0], 'We can only calculate one to size of matrix eigenvalues.'
    assert alphas.ndim == 1, 'The alphas need to be vectors.'
    assert betas.ndim == 1, 'The betas need to be vectors.'
    assert alphas.shape[0] - 1 == betas.shape[0], 'Alpha size needs to be exactly one bigger than beta size.'

    # compute the decomposition
    eigenvalues, eigenvectors = splg.eigh_tridiagonal(d=alphas, e=betas, select='i',
                                                      select_range=(alphas.shape[0] - amount, alphas.shape[0] - 1))

    # return them to be in sinking order
    return eigenvalues[::-1], eigenvectors[:, ::-1]


def rayleigh_ritz_singular_value_decomposition(a_matrix: np.ndarray, k: int) -> (np.ndarray, np.ndarray):
    """
    This function uses the Rayleigh-Ritz method implemented in ARPACK to compute the k highest eigenvalues and
    corresponding eigenvectors. It should be faster than a complete svd.

    !NOTE!:
    The order of the k highest eigenvalues is not guaranteed by this method!

    :param a_matrix: 2D-Matrix filled with floats for which we want to find the left eigenvectors
    :param k: the number of highest eigenvectors we want to find
    :return: returns the eigenvalues and eigenvectors as numpy arrays
    """
    singular_vectors, singular_values, _ = scipy.sparse.linalg.svds(a_matrix, k=k)
    return singular_values, singular_vectors


def facebook_randomized_svd(a_matrix: np.ndarray, randomized_rank: int) -> (np.ndarray, np.ndarray):
    """
    This function implements randomized singular vector decomposition of a matrix as surveyed and described in

    Halko, Nathan, Per-Gunnar Martinsson, and Joel A. Tropp.
    "Finding structure with randomness: Probabilistic algorithms for constructing approximate matrix decompositions."
    SIAM review 53.2 (2011): 217-288.
    on page 4 chapter 1.3 and further.

    :param a_matrix: 2D-Matrix filled with floats for which we want to find the left eigenvectors
    :param randomized_rank: the rank of the noise matrix used for randomized svd. the higher the rank, the better the
    approximation but the lower the precision of the eigenvectors
    :return:
    """
    singular_vectors, singular_values, _ = fbpca.pca(a_matrix, randomized_rank, True)
    return singular_values, singular_vectors


def randomized_hankel_svd(hankel_matrix: np.ndarray, k: int, subspace_iteration_q: int = 2, oversampling_p: int = 2):
    """
    Function for the randomized singular vector decomposition using [1].
    Implementation modified from: https://pypi.org/project/fbpca/
    """

    # get the parameter l from the paper
    sample_length_l = k+oversampling_p
    assert 1.25*sample_length_l < min(hankel_matrix.shape)

    # Apply A to a random matrix, obtaining Q.
    random_matrix_omega = np.random.uniform(low=-1, high=1, size=(hankel_matrix.shape[0], sample_length_l))
    projection_matrix_q = hankel_matrix@random_matrix_omega

    # Form a matrix Q whose columns constitute a well-conditioned basis for the columns of the earlier Q.
    if subspace_iteration_q == 0:
        (projection_matrix_q, _) = sp.linalg.qr(projection_matrix_q, mode='economic')
    if subspace_iteration_q > 0:
        (projection_matrix_q, _) = sp.linalg.lu(projection_matrix_q, permute_l=True)

    # Conduct normalized power iterations.
    for it in range(subspace_iteration_q):

        # QA
        projection_matrix_q = (projection_matrix_q.T @ hankel_matrix).T

        (projection_matrix_q, _) = splg.lu(projection_matrix_q, permute_l=True)

        # AAQ
        projection_matrix_q = hankel_matrix @ projection_matrix_q

        if it + 1 < subspace_iteration_q:
            (projection_matrix_q, _) = splg.lu(projection_matrix_q, permute_l=True)
        else:
            (projection_matrix_q, _) = splg.qr(projection_matrix_q, mode='economic')

    # SVD Q'*A to obtain approximations to the singular values and right singular vectors of A; adjust the left singular
    # vectors of Q'*A to approximate the left singular vectors of A.
    lower_space_hankel = projection_matrix_q.T @ hankel_matrix
    (R, s, Va) = splg.svd(lower_space_hankel, full_matrices=False)
    U = projection_matrix_q.dot(R)

    # Retain only the leftmost k columns of U, the uppermost k rows of Va, and the first k entries of s.
    return U[:, :k], s[:k], Va[:k, :]


@nb.jit(nopython=True)
def compile_hankel(time_series: np.ndarray, end_index: int, window_size: int, rank: int, lag: int = 1) -> np.ndarray:
    """
    This function constructs a hankel matrix from a 1D time series. Please make sure constructing the matrix with
    the given parameters (end index, window size, etc.) is possible, as this function does no checks due to
    performance reasons.

    :param time_series: 1D array with float values as the time series
    :param end_index: the index (point in time) where the time series starts
    :param window_size: the size of the windows cut from the time series
    :param rank: the amount of time series in the matrix
    :param lag: the lag between the time series of the different columns
    :return: The hankel matrix with lag one
    """

    # make an empty matrix to place the values
    #
    # almost no faster way:
    # https://stackoverflow.com/questions/71410927/vectorized-way-to-construct-a-block-hankel-matrix-in-numpy-or-scipy
    hankel = np.empty((window_size, rank))

    # go through the time series and make the hankel matrix
    for cx in range(rank):
        hankel[:, -cx - 1] = time_series[(end_index - window_size - cx * lag):(end_index - cx * lag)]
    return hankel


@nb.njit(parallel=True)
def fast_numba_hankel_matmul(hankel_fft: np.ndarray, l_windows: int, fft_shape: int, other_matrix: np.ndarray,
                             lag: int):
    # get the shape of the other matrix
    m, n = other_matrix.shape

    # make fft of x (while padding x with zeros) to make up for the lag
    if lag > 1:
        out = np.zeros((lag * m - lag + 1, n), dtype=other_matrix.dtype)
        out[::lag, :] = other_matrix
    else:
        out = other_matrix

    # create the new empty matrix that we will fill with values
    result_buffer = np.empty((l_windows, n))

    # flip the other matrix
    out = np.flipud(out)

    # make a numba parallel loop over the vector of the other matrix (columns)
    for index in nb.prange(n):
        # compute the fft of the vector
        fft_x = sp.fft.rfft(out[:, index], n=fft_shape)

        # multiply the ffts with each other to do the convolution in frequency domain and convert it back
        # and save it into the output buffer
        result_buffer[:, index] = sp.fft.irfft(hankel_fft * fft_x, n=fft_shape)[(m - 1) * lag:(m - 1) * lag + l_windows]
    return result_buffer


@nb.njit(parallel=True)
def fast_numba_hankel_left_matmul(hankel_fft: np.ndarray, n_windows: int, fft_shape: int, other_matrix: np.ndarray,
                                  lag: int):
    # transpose the other matrix
    other_matrix = other_matrix.T

    # get the shape of the other matrix
    m, n = other_matrix.shape

    # create the new empty matrix that we will fill with values
    result_buffer = np.empty((n_windows, n))

    # flip the other matrix
    other_matrix = np.flipud(other_matrix)

    # make a numba parallel loop over the vector of the other matrix (columns)
    for index in nb.prange(n):
        # compute the fft of the vector
        fft_x = sp.fft.rfft(other_matrix[:, index], n=fft_shape)

        # multiply the ffts with each other to do the convolution in frequency domain and convert it back
        # and save it into the output buffer
        result_buffer[:, index] = sp.fft.irfft(hankel_fft * fft_x, n=fft_shape)[(m - 1):(m - 1) + n_windows * lag:lag]
    return result_buffer.T


@nb.njit(parallel=True)
def fast_numba_hankel_correlation_matmul(hankel_fft: np.ndarray, fft_shape: int, window_number: int,
                                         other_matrix: np.ndarray, lag: int):

    # get the shape of the other matrix
    m, n = other_matrix.shape
    wn = window_number

    # make a buffer if we need it for lagged representation
    if lag > 1:
        buffer = np.zeros((lag * wn - lag + 1, n), dtype=other_matrix.dtype)
    else:
        # we can just use the other matrix, as we won't ever touch the buffer for lag < 1
        # so we don't modify it
        buffer = other_matrix

    # create the new empty matrix that we will fill with values
    result_buffer = np.empty((m, n))

    # flip the other matrix
    out = np.flipud(other_matrix)

    # make a numba parallel loop over the vector of the other matrix (columns)
    for index in nb.prange(n):

        # compute the fft of the vector
        fft_x = sp.fft.rfft(out[:, index], n=fft_shape)

        # make the convolution and transform it back
        first = sp.fft.irfft(hankel_fft * fft_x, n=fft_shape)[(m - 1):(m - 1) + lag * wn:lag]
        # assert np.allclose(first, result[:, index]), f'{first}\n{result[:, index]}'

        # check whether we have lag
        if lag > 1:
            buffer[::lag, index] = first
            first = buffer[:, index]
        first = np.flip(first)

        # compute the fft of the vector
        fft_x = sp.fft.rfft(first, n=fft_shape)

        # multiply the ffts with each other to do the convolution in frequency domain and convert it back
        # and save it into the output buffer
        result_buffer[:, index] = sp.fft.irfft(hankel_fft*fft_x, n=fft_shape)[(wn-1)*lag:(wn-1)*lag+m]
    return result_buffer


def get_fast_hankel_representation(time_series, end_index, length_windows, number_windows,
                                   lag=1) -> (np.ndarray, int, np.ndarray):
    # get the last column of the hankel matrix. The reason for that is that we will use an algorithm for Toeplitz
    # matrices to speed up the multiplication and Hankel[:, ::-1] == Toeplitz.
    #
    # The algorithm requires the first Toeplitz column and therefore the last Hankel Column
    #
    # It also requires the inverse of the row columns ignoring the last element. For reference see:
    # Scipy: https://docs.scipy.org/doc/scipy/reference/generated/scipy.linalg.matmul_toeplitz.html
    # Pyroomacoustics: https://github.com/LCAV/pyroomacoustics/blob/master/pyroomacoustics/adaptive/util.py
    #
    # additionally we only need the combined vector from both for our multiplication. In order to employ the
    # fastest DFT possible, the vector needs have a length with a clean power of two, which we create using
    # zero padding
    #
    # We can also use the signal itself, as multiplying a vector with the hankel matrix is effectively a convolution
    # over the signal. Therefore, we can use the fft convolution with way lower complexity or any other built in
    # convolution library!

    # get column and row, or we can also do it by using the signal itself!
    # last_column = time_series[end_index-length_windows:end_index]
    # row_without_last_element = time_series[end_index-lag*(number_windows-1)-length_windows:end_index-length_windows]
    signal = time_series[end_index - lag * (number_windows - 1) - length_windows:end_index]

    # get the length of the matrices
    # col_length = last_column.shape[0]
    # row_length = row_without_last_element.shape[0]
    # combined_length = col_length + row_length

    # compute the padded vector length for an optimal fft length. Here we can use the built-in scipy function that
    # computes the perfect fft length optimized for their implementation, so it is even faster than with powers
    # of two!
    # fft_len = 1 << int(np.ceil(np.log2(combined_length)))
    # fft_len = sp.fft.next_fast_len(combined_length, True)
    fft_len = sp.fft.next_fast_len(signal.shape[0] + number_windows, True)

    # compute the fft over the padded hankel matrix
    # if we would pad in the middle like this: we would introduce no linear phase to the fft
    #
    # padded_toeplitz = np.concatenate((last_column, np.zeros((fft_len - combined_length)), row_without_last_element))
    #
    # but we want to use the built-in padding functionality of the fft in scipy, so we pad at the end like can be seen
    # here.
    # This introduces a linear phase and a shift to the fft, which we need to account for in the reverse
    # functions.
    #
    # More details see:
    # https://dsp.stackexchange.com/questions/82273/why-to-pad-zeros-at-the-middle-of-sequence-instead-at-the-end-of-the-sequence
    # https://dsp.stackexchange.com/questions/83461/phase-of-an-fft-after-zeropadding
    #
    # Workers are not necessary as we expect a 1D time series
    hankel_rfft = sp.fft.rfft(signal, n=fft_len, axis=0).reshape(-1, 1)
    return hankel_rfft[:, 0], fft_len, signal


class HankelFFTRepresentation:

    def __init__(self, time_series: np.ndarray, end_index: int, window_length: int, window_number: int, lag: int = 1,
                 _copy_representation: 'HankelFFTRepresentation' = None):

        if _copy_representation is None:

            # save the parameters that are necessary for later computations
            self.window_length = window_length
            self.window_number = window_number
            self.lag = lag
            self.shape = (window_length, window_number)

            # create the representation and save it into the class
            hankel_rfft, fft_len, _ = get_fast_hankel_representation(time_series, end_index,
                                                                     window_length, window_number, lag=lag)
            self.hankel_rfft = hankel_rfft
            self.fft_length = fft_len
        else:

            # check that nothing else is set
            assert time_series is None and end_index is None and window_length is None and lag is None \
                , 'Using the constructor for copying requires all other options to be set to None.'

            # copy the parameters from the other model
            self.window_length = _copy_representation.window_length
            self.window_number = _copy_representation.window_number
            self.lag = 1
            self.hankel_rfft = _copy_representation.hankel_rfft
            self.fft_length = _copy_representation.fft_length

    def multiply_other_from_right(self, other_matrix: np.ndarray) -> np.ndarray:
        # check the dimensions
        if other_matrix.shape[0] != self.window_number:
            raise ValueError(f'matmul: Right matrix has a mismatch in its core dimension 0 '
                             f'(size {other_matrix.shape[0]} is different from {self.window_number})')

        # make the product
        return fast_numba_hankel_matmul(self.hankel_rfft, self.window_length, self.fft_length, other_matrix,
                                        self.lag)

    def multiply_other_from_left(self, other_matrix: np.ndarray) -> np.ndarray:

        # check the dimensions
        if other_matrix.shape[1] != self.window_length:
            raise ValueError(f'matmul: Left matrix has a mismatch in its core dimension 0 '
                             f'(size {other_matrix.shape[1]} is different from {self.window_length})')

        # make the product
        return fast_numba_hankel_left_matmul(self.hankel_rfft, self.window_number, self.fft_length, other_matrix,
                                             self.lag)

    @property
    def T(self) -> 'HankelFFTRepresentation':

        # create a shallow copy
        new_matrix = self.shallow_copy()

        # check that we have no lag in column direction, as we have not implemented lag in row direction
        # for the multiplications yet
        if new_matrix.lag != 1:
            raise ValueError('As of now we can not calculate with transposed lagged Hankel matrices.')

        # switch the window sizes
        new_matrix.window_number, new_matrix.window_length = new_matrix.window_length, new_matrix.window_number
        return new_matrix

    def __matmul__(self, other):

        # right side matmul
        if isinstance(other, np.ndarray):
            return self.multiply_other_from_right(other)

        # right side matmul with transposed
        elif isinstance(other, self.__class__):
            # check that the fft representation is the same and only transposed
            if self.equal_hankel_transposed(other):
                return HankelCorrelationFFTRepresentation(self)
            else:
                raise ValueError('At this point matmul is only supported for itself with its transposed from left.')

    def __array_ufunc__(self, ufunc, method, *args, **kwargs):
        if method != '__call__' or len(kwargs) > 0 or len(args) != 2:
            raise NotImplemented
        if ufunc == np.matmul:
            # left side matmul
            if isinstance(args[0], np.ndarray) and args[1] is self:
                return self.multiply_other_from_left(args[0])
            else:
                raise NotImplemented
        else:
            raise NotImplemented

    def shallow_copy(self) -> 'HankelFFTRepresentation':

        # type hints don't like this, but I don't want to add None to the hint, as normal users should not use
        # directly use the constructor for copying.
        return HankelFFTRepresentation(time_series=None, end_index=None, window_length=None, window_number=None,
                                       lag=None, _copy_representation=self)

    def equal_hankel(self, other: 'HankelFFTRepresentation') -> bool:
        return (self.hankel_rfft is other.hankel_rfft
                and self.window_number == other.window_number and self.window_length == other.window_length)

    def equal_hankel_transposed(self, other: 'HankelFFTRepresentation') -> bool:
        return (self.hankel_rfft is other.hankel_rfft
                and self.window_number == other.window_length and self.window_length == other.window_number)


class HankelCorrelationFFTRepresentation:

    def __init__(self, hankel_matrix: HankelFFTRepresentation):

        # save the parameters that are necessary for later computations
        self.hankel_rfft = hankel_matrix.hankel_rfft
        self.fft_length = hankel_matrix.fft_length
        self.window_length = hankel_matrix.window_length
        self.window_number = hankel_matrix.window_number
        self.lag = hankel_matrix.lag
        self.hankel_matrix = hankel_matrix
        self.shape = (self.window_length, self.window_length)

    def __matmul__(self, other_matrix: np.ndarray) -> np.ndarray:
        if isinstance(other_matrix, np.ndarray):
            return fast_numba_hankel_correlation_matmul(self.hankel_rfft, self.fft_length, self.window_number, other_matrix, self.lag)
        else:
            raise NotImplemented


def examples():
    """
    This function implements some usage examples for quick internal testing. It is not aimed at being used.
    :return: None
    """

    import time

    # create a random signal
    signal = np.random.rand(100)*1000

    # create the two different hankel representations
    hankel_matrix_fft = HankelFFTRepresentation(signal, 100, 50, 20, lag=1)
    hankel_matrix = compile_hankel(signal, end_index=100, window_size=50, rank=20, lag=1)

    # create test matrices to multiply with
    other_matrix = np.random.rand(20, 65)
    other_matrix2 = np.random.rand(50, 15)

    # test the right product
    res = hankel_matrix_fft @ other_matrix
    res2 = hankel_matrix @ other_matrix
    print('Test right:', np.allclose(res, res2))

    # test the left product
    res = other_matrix2.T @ hankel_matrix_fft
    res2 = other_matrix2.T @ hankel_matrix
    print('Test left:', np.allclose(res, res2))

    # test the transposed right product
    res = hankel_matrix_fft.T @ other_matrix2
    res2 = hankel_matrix.T @ other_matrix2
    print('Test transposed right:', np.allclose(res, res2))

    # test the transposed left product
    res = other_matrix.T @ hankel_matrix_fft.T
    res2 = other_matrix.T @ hankel_matrix.T
    print('Test transposed left:', np.allclose(res, res2))

    # check the correlation matrix multiplication
    hankel_corr = hankel_matrix @ hankel_matrix.T
    hankel_fft_corr = hankel_matrix_fft @ hankel_matrix_fft.T
    res = hankel_fft_corr @ other_matrix2
    res2 = hankel_corr @ other_matrix2
    print('Test correlation multiplication:', np.allclose(res, res2))

    # set a random seed
    np.random.seed(1234)

    # create the random starting vector
    x = np.random.rand(100)
    x /= np.linalg.norm(x)
    A = np.random.rand(100, 100)

    # test the power method
    singval, singvec = power_method(A, x, n_iterations=100)
    singvecs, singvals, _ = np.linalg.svd(A)
    print(singval, singvals[0])

    # test the tridiagonalization method for speed
    size = 1000
    d = 3 * np.random.rand(size)
    e = -1 * np.random.rand(size - 1)
    start = time.time()
    tri_eigvals, tri_eigvecs = tridiagonal_eigenvalues(d, e, size // 2)
    print(f'Specialized tridiagonal SVD took: {time.time() - start} s.')
    T = np.diag(d) + np.diag(e, k=1) + np.diag(e, k=-1)
    start = time.time()
    singvecs, singvals, _ = np.linalg.svd(T)
    print(f'Normal SVD took: {time.time() - start} s.')

    # test randomized svd
    singval, singvec = facebook_randomized_svd(A, 20)
    singvecs, singvals, _ = np.linalg.svd(A)
    print(singval, singvals[0])


if __name__ == '__main__':
    examples()



import numpy as np
from typing import Callable
from functools import partial

import torch

from changepoynt.algorithms.base_algorithm import Algorithm
from changepoynt.utils import normalization
from changepoynt.utils import linalg as lg


class TESST(Algorithm):
    """
    This class implements an own idea for change point detection.

    It is not yet published, please use with caution and do not copy or use in production (Lucas Weber, 2023).
    """

    def __init__(self, window_length: int, n_windows: int = None, lag: int = None, rank: int = 5, scoring_step: int = 1,
                 scale: bool = True) -> None:
        """
        We initialize the matrix profile and the subsequent floss using only the window length used for comarisons.

        :param window_length: the length of the window used for the distance comparisons in the matrix profile.
        """

        # save the specified parameters into instance variables
        self.window_length = window_length
        self.n_windows = n_windows
        self.rank = rank
        self.scale = scale
        self.lag = lag
        self.scoring_step = scoring_step

        # set some default values when they have not been specified
        if self.n_windows is None:
            self.n_windows = self.window_length//2
        if self.lag is None:
            # rule of thumb
            self.lag = self.n_windows

        # check whether we have cuda
        assert torch.cuda.is_available(), 'A torch installation with cuda is necessary for this.'

        # specify the methods and their corresponding functions as lambda functions expecting only the hanke matrix
        self.methods = {'rsvd': partial(left_entropy, rank=self.rank)}

    def transform(self, time_series: np.ndarray) -> np.ndarray:
        """
        This function calculate the anomaly score for each sample within the time series.

        It also does some assertion regarding time series length.

        :param time_series: 1D array containing the time series to be scored
        :return: anomaly score
        """

        # check the dimensions of the input array
        assert time_series.ndim == 1, "Time series needs to be an 1D array."

        # check that we have at least two windows
        assert time_series.shape[0] > self.window_length, 'Time series needs to be longer than window length.'

        # compute the starting point of the scoring (past and future hankel need to fit)
        starting_point = self.window_length + self.n_windows + self.lag
        assert starting_point < time_series.shape[0], "The time series is too short to score any points."

        # scale the time series (or just copy it if already scaled)
        if self.scale:
            time_series = normalization.min_max_scaling(time_series, min_val=1.0, max_val=2.0, inplace=False)
        else:
            time_series = time_series.copy()

        return _transform(time_series, starting_point, self.window_length, self.n_windows, self.lag, self.scoring_step,
                          self.methods['rsvd'])


def _transform(time_series: np.ndarray, start_idx: int, window_length: int, n_windows: int, lag: int, scoring_step: int,
               scoring_function: Callable) -> np.ndarray:
    """
    Compute heavy and hopefully jit compilable score computation for the SST method. It does not do any parameter
    checking and can throw cryptic errors. It's only used for internal use as a private function.

    :param time_series: 1D array containing the time series to be scored
    :param start_idx: integer defining the start sample index for the score computation
    :param window_length: the size of the time series window for each column of the hankel matrix
    :param n_windows: amount of columns in the hankel matrix
    """

    # initialize a scoring array with no values yet
    score = np.zeros_like(time_series)

    # compute the offset
    offset = (n_windows + lag)

    # compute the batched hankel matrix
    batched_hankel = np.zeros(((time_series.shape[0]-start_idx)//scoring_step, window_length, 2*n_windows))

    # iterate over all the values in the signal and construct the hankel matrix
    for idx in range(start_idx, time_series.shape[0], scoring_step):

        # compile the past hankel matrix (H1)
        hankel_past = lg.compile_hankel(time_series, idx - lag, window_length, n_windows)

        # compile the future hankel matrix (H2)
        hankel_future = lg.compile_hankel(time_series, idx, window_length, n_windows)

        # put the hankel matrix into the batched hankel matrix
        batched_hankel[idx-start_idx, :, :] = np.concatenate((hankel_past, hankel_future), axis=1)

    # transform into and torch tensor and put onto the gpu
    with torch.no_grad():
        batched_hankel = torch.from_numpy(batched_hankel).to('cuda:0')
        print(batched_hankel.shape)
        print(f'Tensor is on the GPU: {torch.cuda.get_device_name(0)}')

        # compute the score
        import time
        start = time.time()
        cuda_score = scoring_function(batched_hankel)
        print(f'Computation on GPU took {time.time() - start} s.')

        # repeat the numbers and get back to cpu
        cuda_score = cuda_score.repeat_interleave(scoring_step).cpu().numpy()
        score[start_idx-offset-scoring_step//2:-offset] = cuda_score

    return score


def left_right_diff(left_eigenvectors: torch.Tensor):
    n = left_eigenvectors.shape[1]//2
    return torch.mean(left_eigenvectors[:, :n, :]-left_eigenvectors[:, n:, :], dim=1)


def left_entropy(hankel: torch.Tensor, rank: int) -> float:
    """
    Entropy Singular Spectrum Transformation.

    :param hankel: the hankel matrix of the signal
    :param rank: the amount of (approximated) eigenvectors as subspace of H1
    :return: the change point score, the input vector x0
    """
    # compute the left and right eigenvectors using the randomized svd for fast computation
    left_eigenvectors, eigenvalues, right_eigenvectors = torch.svd_lowrank(hankel, rank)

    # compute the difference between left and right participation of eigenvectors
    difference = torch.abs(left_right_diff(right_eigenvectors))

    # compute the weighted mean of the difference
    weighted_difference = torch.sum(difference*eigenvalues, dim=1) / torch.sum(eigenvalues)

    return weighted_difference


def _main():
    """
    Internal quick testing function.

    :return:
    """
    from time import time
    # make synthetic step function
    np.random.seed(123)
    # synthetic (frequency change)
    x0 = np.sin(2 * np.pi * 1 * np.linspace(0, 10, 1000))
    x1 = np.sin(2 * np.pi * 2 * np.linspace(0, 10, 1000))
    x2 = np.sin(2 * np.pi * 8 * np.linspace(0, 10, 1000))
    x3 = np.sin(2 * np.pi * 4 * np.linspace(0, 10, 1000))
    x = np.hstack([x0, x1, x2, x3])
    x += np.random.rand(x.size)

    # create the method
    esst_recognizer = TESST(50)

    # compute the score
    start = time()
    score = esst_recognizer.transform(x)
    print(f'Computation for {len(x)} signal values took {time()-start} s.')


if __name__ == '__main__':
    _main()



import pytest
import numpy as np
import changepoynt.algorithms.si as si
import logging
logging.basicConfig(level=logging.DEBUG)
LOGGER = logging.getLogger(__name__)


class TestSST:
    def setup_method(self):
        # set a random seed
        np.random.seed(3455)

        # create a random steps signal of a certain length
        self.signal_length = 200
        x0 = 1 * np.ones(self.signal_length) + np.random.rand(self.signal_length) * 1
        x1 = 3 * np.ones(self.signal_length) + np.random.rand(self.signal_length) * 2
        x2 = 5 * np.ones(self.signal_length) + np.random.rand(self.signal_length) * 1.5
        x = np.hstack([x0, x1, x2])
        x += np.random.rand(x.size)
        self.signal = x

    def teardown_method(self):
        pass

    def test_default(self):
        with pytest.raises(NotImplementedError):
            si.SubspaceIdentification(20).transform(self.signal)


if __name__ == "__main__":
    pytest.main()



import numpy as np
from changepoynt.algorithms.base_algorithm import Algorithm
# from claspy.segmentation import BinaryClaSPSegmentation


class CLASP(Algorithm):
    """
    This class uses the ClaSP algorithm published in

    [1]
    Schäfer, Patrick, Arik Ermshaus, and Ulf Leser.
    "Clasp-time series segmentation."
    Proceedings of the 30th ACM International Conference on Information & Knowledge Management. 2021.

    This class essentially wraps the claspy library
    https://github.com/ermshaua/claspy
    https://sites.google.com/view/ts-clasp/

    In the current configuration it is recommended to set no parameters, as CLASP is able to extract it from the data.
    TODO:
    Own implementation?
    """

    def __init__(self, n_segments="learn", n_estimators=10, window_size="suss", k_neighbours=3,
                 distance="znormed_euclidean_distance", score="roc_auc",
                 early_stopping=True, validation="significance_test", threshold=1e-15, excl_radius=5,
                 random_state=2357) -> None:

        # say that we currently not support CLASP due to old packages in the requirements
        raise NotImplementedError('CLASP is currently not available, as it requires outdated package versions.')

        # save the specified parameters into instance variables
        self.n_segments = n_segments
        self.n_estimators = n_estimators
        self.window_size = window_size
        self.k_neighbours = k_neighbours
        self.distance = distance
        self.score = score
        self.early_stopping = early_stopping
        self.validation = validation
        self.threshold = threshold
        self.excl_radius = excl_radius
        self.random_state = random_state

    def transform(self, time_series: np.ndarray) -> np.ndarray:
        # check the dimensions of the input array
        assert time_series.ndim == 1, "Time series needs to be an 1D array."

        # make the clasp segmentation object
        claspy_segmenter = BinaryClaSPSegmentation(n_segments=self.n_segments, n_estimators=self.n_estimators,
                                                   window_size=self.window_size, k_neighbours=self.k_neighbours,
                                                   score=self.score, early_stopping=self.early_stopping,
                                                   validation=self.validation, threshold=self.threshold,
                                                   excl_radius=self.excl_radius, random_state=self.random_state)
        claspy_segmenter.fit(time_series)
        return claspy_segmenter.profile


def _main():
    """
    Internal quick testing function.

    :return:
    """
    from time import time
    # make synthetic step function
    np.random.seed(123)
    # synthetic (frequency change)
    x0 = np.sin(2 * np.pi * 1 * np.linspace(0, 10, 1000))
    x1 = np.sin(2 * np.pi * 2 * np.linspace(0, 10, 1000))
    x2 = np.sin(2 * np.pi * 8 * np.linspace(0, 10, 1000))
    x3 = np.sin(2 * np.pi * 4 * np.linspace(0, 10, 1000))
    x = np.hstack([x0, x1, x2, x3])
    x += np.random.rand(x.size)

    # create the method
    clasp_recognizer = CLASP()

    # compute the score
    start = time()
    score = clasp_recognizer.transform(x)
    print(f'Computation for {len(x)} signal values took {time()-start} s.')


if __name__ == '__main__':
    _main()



import numpy as np
from changepoynt.algorithms.base_algorithm import Algorithm
import stumpy


class FLOSS(Algorithm):
    """
    This class uses the FLOSS algorithm described in:

    [1]
    Gharghabi, Shaghayegh, et al.
    "Matrix profile VIII: domain agnostic online semantic segmentation at superhuman performance levels."
    2017 IEEE international conference on data mining (ICDM). IEEE, 2017.

    This class essentially wraps the stumpy library which implements a fast approach to calculate the matrix profile.
    https://stumpy.readthedocs.io/en/latest/index.html

    TODO:
    Implement GPU support and streaming? (FLOSS)
    """

    def __init__(self, window_length: int, initial_length: int = None) -> None:
        """
        We initialize the matrix profile and the subsequent floss using only the window length used for comarisons.

        :param window_length: the length of the window used for the distance comparisons in the matrix profile.
        :param initial_length: the length for the initial matrix profile length
        """
        raise NotImplementedError('FLOSS is not yet fully functional.')
        # save the specified parameters into instance variables
        self.window_length = window_length
        self.initial_length = initial_length

        # set default initial length if necessary
        if not initial_length:
            self.initial_length = 10*self.window_length

        # check for the size
        assert self.initial_length >= 2*self.window_length, 'The initial_length should be at least twice' \
                                                            'the window_length.'

    def transform(self, time_series: np.ndarray) -> np.ndarray:
        """
        This function calculate the anomaly score for each sample within the time series.

        It also does some assertion regarding time series length.

        :param time_series: 1D array containing the time series to be scored
        :return: anomaly score
        """

        # check the dimensions of the input array
        assert time_series.ndim == 1, "Time series needs to be an 1D array."

        # check that we have at least two windows
        assert time_series.shape[0] > self.initial_length, \
            'Time series needs to be longer than specified initial length.'

        # feed it through the online process
        score = _transform(time_series, self.initial_length, self.window_length)
        return score


def save_animation(mp, windows, time_series):
    import matplotlib.pyplot as plt
    from matplotlib import animation
    import os

    fig, axs = plt.subplots(2, sharex=True, gridspec_kw={'hspace': 0})

    axs[0].set_xlim((0, mp.shape[0]))
    axs[0].set_ylim((min(time_series), max(time_series)))
    axs[1].set_xlim((0, mp.shape[0]))
    axs[1].set_ylim((-0.1, 1.1))

    lines = []
    for ax in axs:
        line, = ax.plot([], [], lw=2)
        lines.append(line)
    line, = axs[1].plot([], [], lw=2)
    lines.append(line)

    # mark the window lengths
    axs[1].plot([401, 401], (0, mp.shape[0]))

    def init():
        for line in lines:
            line.set_data([], [])
        return lines

    def animate(window):
        data_out, cac_out = window
        for line, data in zip(lines, [data_out, cac_out]):
            line.set_data(np.arange(data.shape[0]), data)
        return lines

    anim = animation.FuncAnimation(fig, animate, init_func=init,
                                   frames=windows, interval=100,
                                   blit=True)
    writergif = animation.PillowWriter(fps=5)
    anim.save('semantic.gif', writer=writergif)


def _transform(time_series: np.ndarray, start_idx: int, window_length: int) -> np.ndarray:
    """
    Compute FLOSS from [1] as the inverse online 1D corrected arc crossing rate.

    :param time_series: 1D array containing the time series to be scored
    :param start_idx: the size for the initial matrix profile
    :param window_length: the size of the windows to be compared
    """

    # create the initial signal
    init_signal = time_series[:start_idx]

    # create the initial matrix profile
    matrix_profile = stumpy.stump(init_signal, m=window_length)

    # initialize the floss object
    stream = stumpy.floss(matrix_profile, init_signal, m=window_length, L=window_length, excl_factor=1)

    # make the score vector
    score = np.zeros_like(time_series)
    score[:start_idx] = 1

    # iterate over all the values in the signal starting at start_idx computing the change point score
    windows = []
    for idx in range(start_idx, time_series.shape[0]):

        # update the floss streaming module
        stream.update(time_series[idx])

        # get the latest score (1-cac)
        score[idx] = stream.cac_1d_[-window_length-2]

        if idx % 20 == 0: windows.append((stream.T_, stream.cac_1d_))
    save_animation(matrix_profile, windows, time_series)
    return 1-score


def _main():
    """
    Internal quick testing function.

    :return:
    """
    from time import time
    import matplotlib.pyplot as plt
    # make synthetic step function
    np.random.seed(123)
    # synthetic (frequency change)
    x0 = np.sin(2 * np.pi * 1 * np.linspace(0, 10, 1000))
    x1 = np.sin(2 * np.pi * 2 * np.linspace(0, 10, 1000))
    x2 = np.sin(2 * np.pi * 8 * np.linspace(0, 10, 1000))
    x3 = np.sin(2 * np.pi * 4 * np.linspace(0, 10, 1000))
    x = np.hstack([x0, x1, x2, x3])
    x += np.random.rand(x.size)

    # create the method
    fluss_recognizer = FLOSS(50)

    # compute the score
    start = time()
    score = fluss_recognizer.transform(x)
    plt.plot(x)
    plt.plot(score)
    plt.show()
    print(f'Computation for {len(x)} signal values took {time()-start} s.')


if __name__ == '__main__':
    _main()



import numpy as np
from changepoynt.utils import linalg as lg
from changepoynt.utils import densityratioestimation as dre
from changepoynt.algorithms.base_algorithm import Algorithm
import multiprocessing as mp


class RuLSIF(Algorithm):
    """
    This class implements change point detection based on relative density ration estimation optimizing the least
    squares approach (RuLSIF) from:

    [1]
    Liu, Song, et al.
    "Change-point detection in time-series data by relative density-ratio estimation."
    Neural Networks 43 (2013): 72-83.

    The code has been created looking at:
    - https://github.com/hoxo-m/densratio_py
    - https://pypi.org/project/densratio/
    - http://www.makotoyamada-ml.com/RuLSIF.html (Matlab implementation original author)
    - https://github.com/anewgithubname/change_detection (Matlab implementation co-author)
    """

    def __init__(self, window_length: int = 10, n_windows: int = 50, lag: int = None, estimation_lag: int = None,
                 scoring_step: int = 1, n_kernels: int = 100, alpha: float = 0.01, symmetric=True) -> None:
        """
        This defines all necessary parameters for the RuLSIF to work.
        :param window_length: the length of the windows we want to compare the densities for (k in the paper)
        :param n_windows: the amount of windows (n in the paper)
        :param lag: the difference in time between the past and the future comparison window
        :param estimation_lag: how often we should re-estimate lambda and sigma for the gaussian kernels
        :param scoring_step: the number of samples between each change score value (e.g. 2 would half the computations)
        :param n_kernels: the number of kernels for the density ration estimation
        :param alpha: the smoothing parameter for the RELATIVE in RuLSIF
        :param symmetric: specifies whether to use two processes to compute a forward and backward pass simultaneously
        """
        self.window_length = window_length
        self.n_windows = n_windows
        self.lag = lag
        self.estimation_lag = estimation_lag
        self.n_kernels = n_kernels
        self.alpha = alpha
        self.scoring_step = scoring_step
        self.symmetric = symmetric

        # check that alpha does not exceed the bounds
        assert 0 <= self.alpha < 1, 'The alpha parameter should be in the interval [0,1).'

        # check for the estimation lag
        assert self.estimation_lag is None or 1 <= self.estimation_lag, \
            'The estimation lag needs to be bigger than zero samples.'

        # set the lag if it is not given
        if not self.lag:
            self.lag = self.n_windows

    def transform(self, time_series: np.ndarray):

        # check the dimensions of the input array
        assert time_series.ndim == 1, "Time series needs to be an 1D array."

        # compute the starting point of the scoring (past and future hankel need to fit)
        starting_point = self.window_length + self.n_windows + self.lag
        assert starting_point < time_series.shape[0], "The time series is too short to score any points."

        # create the estimator
        estimator = dre.RuLSIF(self.alpha)

        # copy the time series to protect it from any modifications within the algorithm
        time_series = time_series.copy()

        # check whether we want to compute the symmetric score
        if self.symmetric:

            # compute the scores with two workers one going forward and the other one backward
            with mp.Pool(2) as workers:
                result = workers.starmap(_transform,
                                         ((time_series, starting_point, self.window_length, self.n_windows, self.lag,
                                           self.scoring_step, estimator),
                                          (time_series[::-1], starting_point, self.window_length, self.n_windows,
                                           self.lag, self.scoring_step, estimator)))

            return result[0] + result[1][::-1]

        else:
            # call the function to compute the values
            return _transform(time_series, starting_point, self.window_length, self.n_windows, self.lag,
                              self.scoring_step, estimator)


def _transform(time_series: np.ndarray, starting_point: int, window_length: int, n_windows: int, lag: int,
               scoring_step: int, estimator: dre.RuLSIF) -> np.ndarray:

    # create the empty score vector
    score = np.zeros_like(time_series)

    # compute the offset
    offset = n_windows

    # iterate over all the values in the signal starting at start_idx computing the change point score
    for idx in range(starting_point, time_series.shape[0], scoring_step):

        # compile the past hankel matrix (Y)
        hankel_matrix = lg.compile_hankel(time_series, idx, window_length, 2*n_windows)

        # compute the score
        score[idx-offset-scoring_step//2:idx-offset+(scoring_step+1)//2] = estimator(hankel_matrix[:, :n_windows],
                                                                                     hankel_matrix[:, n_windows:])

    return score


def short_test():
    from time import time
    import matplotlib.pyplot as plt
    import scipy

    # make synthetic step function
    np.random.seed(123)
    length = 200
    x = np.hstack([1 * np.ones(length) + np.random.rand(length) * 1,
                   3 * np.ones(length) + np.random.rand(length) * 2,
                   5 * np.ones(length) + np.random.rand(length) * 1.5])
    x = np.hstack([np.random.normal(scale=4, size=length),
                   np.random.normal(scale=1, size=length),
                   np.random.normal(scale=4, size=length)])
    x = scipy.io.loadmat('..\..\examples\logwell.mat')["y"][0, :]

    # plot and show the signal
    plt.plot(x)
    plt.show()

    # create the rulsif method
    ruli = RuLSIF(symmetric=True)

    # make the scoring
    start = time()
    score = ruli.transform(x)
    print((time() - start))
    plt.plot(score)
    plt.show()


if __name__ == '__main__':
    short_test()



import matplotlib.pyplot as plt
import numpy as np
from typing import Callable
from functools import partial
from changepoynt.algorithms.base_algorithm import Algorithm
from changepoynt.utils import normalization
from changepoynt.utils import linalg as lg
import fbpca
import multiprocessing as mp


class ESST(Algorithm):
    """
    This class implements an own idea for change point detection.

    It is not yet published, please use with caution and do not copy or use in production (Lucas Weber, 2023).
    """

    def __init__(self, window_length: int, n_windows: int = None, lag: int = None, rank: int = 5,
                 scale: bool = True, scoring_step: int = 1, parallel=False) -> None:
        """
        Experimental change point detection method evaluation the prevalence of change points within a signal
        by comparing the difference in eigenvectors between to points in time.

        :param window_length: This specifies the length of the time series (in samples), which will be used to extract
        the representative "eigensignals" to be compared before and after the lag. The windows length should be big
        enough to cover any wanted patterns (e.g. bigger than the periodicity of periodic signals).

        :param n_windows: This specifies the amount of consecutive time windows used to extract the "eigensignals" from
        the given time series. It should be big enough to cover different parts of the target behavior. If one does not
        specify this value, we use the rule of thumb and take as many time windows as you specified the length of the
        window (rule of thumb).

        :param lag: This parameter specifies the distance of the comparison for behaviors. In easy terms it tells the
        algorithms how far it should look into the future to find change in behavior to the current signal. If you do
        not specify this parameter, we use a rule of thumb and look ahead half of the window length you specified to
        cover the behavior.

        :param rank: This parameter specifies the amount of "eigensignals" which will be used to measure the
        dissimilarity of the signal in the future behavior. As a rule of thumb we take the five most dominant
        "eigensignals" if you do no specify otherwise.

        :param scale: Due to numeric stability we REALLY RECOMMEND scaling the signal into a restricted value range. Per
        default, we use a min max scaling to ensure a restricted value range. In the presence of extreme outliers this
        could cause problems, as the signal will be squished.

        :param scoring_step: the distance between scoring steps in samples (e.g. 2 would half the computation).

        :param parallel: the execution for the different steps can be parallelized in different processes.
        """

        # save the specified parameters into instance variables
        self.window_length = window_length
        self.n_windows = n_windows
        self.rank = rank
        self.scale = scale
        self.lag = lag
        self.scoring_step = scoring_step
        self.parallel = parallel

        # set some default values when they have not been specified
        if self.n_windows is None:
            self.n_windows = self.window_length//2
        if self.lag is None:
            # rule of thumb
            self.lag = self.n_windows

        # specify the methods and their corresponding functions as lambda functions expecting only the hanke matrix
        self.methods = {'rsvd': partial(left_entropy, rank=self.rank)}

    def transform(self, time_series: np.ndarray) -> np.ndarray:
        """
        This function calculate the anomaly score for each sample within the time series.

        It also does some assertion regarding time series length.

        :param time_series: 1D array containing the time series to be scored
        :return: anomaly score
        """

        # check the dimensions of the input array
        assert time_series.ndim == 1, "Time series needs to be an 1D array."

        # check that we have at least two windows
        assert time_series.shape[0] > self.window_length, 'Time series needs to be longer than window length.'

        # compute the starting point of the scoring (past and future hankel need to fit)
        starting_point = self.window_length + self.n_windows + self.lag
        assert starting_point < time_series.shape[0], "The time series is too short to score any points."

        # scale the time series (or just copy it if already scaled)
        if self.scale:
            time_series = normalization.min_max_scaling(time_series, min_val=1.0, max_val=2.0, inplace=False)
        else:
            time_series = time_series.copy()

        if self.parallel:
            return _transform_parallel(time_series, starting_point, self.window_length, self.n_windows, self.lag,
                                       self.scoring_step, self.methods['rsvd'])
        else:
            return _transform(time_series, starting_point, self.window_length, self.n_windows, self.lag,
                              self.scoring_step, self.methods['rsvd'])


def _transform(time_series: np.ndarray, start_idx: int, window_length: int, n_windows: int, lag: int, scoring_step: int,
               scoring_function: Callable) -> np.ndarray:
    """
    Compute heavy and hopefully jit compilable score computation for the SST method. It does not do any parameter
    checking and can throw cryptic errors. It's only used for internal use as a private function.

    :param time_series: 1D array containing the time series to be scored
    :param start_idx: integer defining the start sample index for the score computation
    :param window_length: the size of the time series window for each column of the hankel matrix
    :param n_windows: amount of columns in the hankel matrix
    """

    # initialize a scoring array with no values yet
    score = np.zeros_like(time_series)

    # compute the offset
    offset = (n_windows + lag)

    # iterate over all the values in the signal starting at start_idx computing the change point score
    for idx in range(start_idx, time_series.shape[0], scoring_step):

        # compile the past hankel matrix (H1)
        hankel_past = lg.compile_hankel(time_series, idx - lag, window_length, n_windows)

        # compile the future hankel matrix (H2)
        hankel_future = lg.compile_hankel(time_series, idx, window_length, n_windows)

        # compute the score and save the returned feedback vector
        score[idx-offset-scoring_step//2:idx-offset+(scoring_step+1)//2] = \
            scoring_function(np.concatenate((hankel_past, hankel_future), axis=1))

    return score


def _transform_parallel(time_series: np.ndarray, start_idx: int, window_length: int, n_windows: int, lag: int,
                        scoring_step: int, scoring_function: Callable) -> np.ndarray:
    """
    Compute heavy and hopefully jit compilable score computation for the SST method. It does not do any parameter
    checking and can throw cryptic errors. It's only used for internal use as a private function.

    :param time_series: 1D array containing the time series to be scored
    :param start_idx: integer defining the start sample index for the score computation
    :param window_length: the size of the time series window for each column of the hankel matrix
    :param n_windows: amount of columns in the hankel matrix
    """

    # initialize a scoring array with no values yet
    score = np.zeros_like(time_series)

    # compute the offset
    offset = (n_windows + lag)

    # make the generator for the hankel matrices
    gener = (np.concatenate(
                            (lg.compile_hankel(time_series, idx-lag, window_length, n_windows),
                             lg.compile_hankel(time_series, idx, window_length, n_windows)
                             ),
                            axis=1)
             for idx in range(start_idx, time_series.shape[0], scoring_step))

    # make a process pool with batches
    with mp.Pool(mp.cpu_count()) as pp:
        for idx, result in enumerate(pp.imap(scoring_function, gener, chunksize=100)):
            idx = start_idx + idx*scoring_step
            score[idx - offset - scoring_step // 2:idx - offset + (scoring_step + 1) // 2] = result

    return score


def skewness(pdf_matrix: np.ndarray, x: np.ndarray) -> np.ndarray:
    """
    Compute the skewness of a matrix of probability density functions given as a numpy array.

    Args:
        pdf_matrix (numpy array): Matrix of probability density functions, where each row is a probability density function
        x (numpy array): Array of values for which the pdfs are defined

    Returns:
        Skewness of each probability density function as a numpy array
    """
    mean_vector = np.sum(x * pdf_matrix, axis=1)
    variances_vector = np.sum((x - mean_vector[:, np.newaxis]) ** 2 * pdf_matrix, axis=1)
    std_vector = np.sqrt(variances_vector)
    skewness_vector = np.sum(((x - mean_vector[:, np.newaxis]) / std_vector[:, np.newaxis]) ** 3 * pdf_matrix, axis=1)
    return skewness_vector


def left_right_diff(left_eigenvectors: np.ndarray):
    n = left_eigenvectors.shape[1]//2
    return np.mean(left_eigenvectors[:, :n]-left_eigenvectors[:, n:], axis=1)


def left_entropy(hankel: np.ndarray, rank: int) -> float:
    """
    Entropy Singular Spectrum Transformation.

    :param hankel: the hankel matrix of the signal
    :param rank: the amount of (approximated) eigenvectors as subspace of H1
    :return: the change point score, the input vector x0
    """
    # compute the left and right eigenvectors using the randomized svd for fast computation
    right_eigenvectors, eigenvalues, left_eigenvectors = fbpca.pca(hankel, rank, True)

    # shift the left eigenvectors up
    left_eigenvectors = left_eigenvectors - np.min(left_eigenvectors, axis=1)[:, None] + 1

    # make the eigenvectors into "probability distributions" so their sum of elements is one
    left_eigenvectors = left_eigenvectors/np.sum(left_eigenvectors, axis=1)[:, None]

    # compute the normalized entropy of the eigenvectors
    """entropy = (-1 *
               np.sum(np.log2(left_eigenvectors, out=np.zeros_like(left_eigenvectors), where=(left_eigenvectors != 0))
                      * left_eigenvectors, axis=1))/np.log2(hankel.shape[1])"""
    # skew = np.abs(skewness(left_eigenvectors, np.tile(np.linspace(0, hankel.shape[1]-1, hankel.shape[1]), (rank, 1))))
    skew = np.abs(left_right_diff(left_eigenvectors))

    # compute the weighted mean of the entropy
    # weighted_entropy = (eigenvalues @ ((1-entropy)*skew))/np.sum(eigenvalues)
    weighted_entropy = (eigenvalues @ skew) / np.sum(eigenvalues)

    return weighted_entropy


def _main():
    """
    Internal quick testing function.

    :return:
    """
    from time import time
    # make synthetic step function
    np.random.seed(123)
    # synthetic (frequency change)
    x0 = np.sin(2 * np.pi * 1 * np.linspace(0, 10, 10000))
    x1 = np.sin(2 * np.pi * 2 * np.linspace(0, 10, 10000))
    x2 = np.sin(2 * np.pi * 8 * np.linspace(0, 10, 10000))
    x3 = np.sin(2 * np.pi * 4 * np.linspace(0, 10, 10000))
    x = np.hstack([x0, x1, x2, x3])
    x += np.random.rand(x.size)

    # create the method
    esst_recognizer = ESST(50, parallel=True)

    # compute the score
    start = time()
    score1 = esst_recognizer.transform(x)
    print(f'Computation for {len(x)} signal values took {time()-start} s.')


if __name__ == '__main__':
    _main()



# this file contains the tests for the linalg.py file in the utils of this package
import pytest
import numpy as np
import changepoynt.utils.linalg as lg


class TestLinearAlgebra:
    def setup_method(self):
        # set a random seed
        np.random.seed(3455)

        # create a random square matrix
        self.A = np.random.rand(50, 50)

        # create random starting vector for the power method
        # and normalize it
        self.x0 = np.random.rand(50)
        self.x0 /= np.linalg.norm(self.x0)

        # create a tridiagonal matrix and the diagonal elements
        self.d = 5 * np.random.rand(100)
        self.e = -1 * np.random.rand(99)
        self.T = np.diag(self.d) + np.diag(self.e, k=1) + np.diag(self.e, k=-1)

        # compute the svd of the matrix A as we will need it for comparisons
        self.singvecs, self.singvals, _ = np.linalg.svd(self.A)

        # create a random signal
        self.signal = np.random.rand(100) * 1000

        # create test matrices to multiply with
        self.other_matrix = np.random.rand(20, 65)
        self.other_matrix2 = np.random.rand(50, 15)

        # create the two different hankel representations
        self.hankel_matrix_fft = lg.HankelFFTRepresentation(self.signal, end_index=100, window_length=50,
                                                            window_number=20, lag=1)
        self.hankel_matrix = lg.compile_hankel(self.signal, end_index=100, window_size=50, rank=20, lag=1)

    def teardown_method(self):
        pass

    def test_power_method(self):
        # make the correlation matrix
        a_corr = self.A @ self.A.T
        eigval, eigvec = lg.power_method(a_corr, self.x0, n_iterations=100)

        # get the eigenvector
        eigvals, eigvecs = np.linalg.eig(a_corr)
        idces = np.argsort(eigvals)[::-1]
        eigvals = eigvals[idces]
        eigvecs = eigvecs[:, idces]

        np.testing.assert_almost_equal(eigval, eigvals[0])
        np.testing.assert_almost_equal(np.abs(eigvec), np.abs(eigvecs[:, 0]))

    def test_eig_tridiag(self):
        # take a look at the highest half of the eigenvalues as lower one might be unstable
        # https://stackoverflow.com/questions/46345217/diagonalization-of-a-tridiagonal-symmetric-sparse-matrix-with-python
        amount = self.d.shape[0]//2
        tri_eigvals, tri_eigvecs = lg.tridiagonal_eigenvalues(self.d, self.e, amount)
        eigvecs, eigvals, _ = np.linalg.svd(self.T)
        np.testing.assert_almost_equal(tri_eigvals, eigvals[:amount])
        np.testing.assert_almost_equal(np.abs(tri_eigvecs), np.abs(eigvecs[:, :amount]))

    def test_rayleigh_ritz_svd(self):
        k = 20
        k_eigvals, k_eigvecs = lg.rayleigh_ritz_singular_value_decomposition(self.A, k)

        # order is not guaranteed by the method, so we need to sort in decreasing order
        idx = np.argsort(k_eigvals)[::-1]
        k_eigvals = k_eigvals[idx]
        k_eigvecs = k_eigvecs[:, idx]

        # compare the values
        np.testing.assert_almost_equal(k_eigvals, self.singvals[:k])
        np.testing.assert_almost_equal(np.abs(k_eigvecs), np.abs(self.singvecs[:, :k]))

    def test_randomized_svd(self):

        # set a threshold for the amount of eigenvectors we want to have and compute them
        k = 3
        k_eigvals, k_eigvecs = lg.facebook_randomized_svd(self.A, 50)

        # compare the values
        np.testing.assert_almost_equal(k_eigvals[:k], self.singvals[:k])
        np.testing.assert_almost_equal(np.abs(k_eigvecs[:, :k]), np.abs(self.singvecs[:, :k]))

    def test_right_hankel_product(self):
        # make the multiplications
        res = self.hankel_matrix_fft @ self.other_matrix
        res2 = self.hankel_matrix @ self.other_matrix
        np.testing.assert_almost_equal(res, res2)

    def test_left_hankel_product(self):
        # make the multiplications
        res = self.other_matrix2.T @ self.hankel_matrix_fft
        res2 = self.other_matrix2.T @ self.hankel_matrix
        np.testing.assert_almost_equal(res, res2)

    def test_transposed_right_hankel_product(self):
        # make the multiplications
        res = self.hankel_matrix_fft.T @ self.other_matrix2
        res2 = self.hankel_matrix.T @ self.other_matrix2
        np.testing.assert_almost_equal(res, res2)

    def test_transposed_left_hankel_product(self):
        # make the multiplications
        res = self.other_matrix.T @ self.hankel_matrix_fft.T
        res2 = self.other_matrix.T @ self.hankel_matrix.T
        np.testing.assert_almost_equal(res, res2)

    def test_hankel_correlation_product(self):
        # check the correlation matrix multiplication
        hankel_corr = self.hankel_matrix @ self.hankel_matrix.T
        hankel_fft_corr = self.hankel_matrix_fft @ self.hankel_matrix_fft.T
        res = hankel_fft_corr @ self.other_matrix2
        res2 = hankel_corr @ self.other_matrix2
        np.testing.assert_almost_equal(res, res2)


if __name__ == "__main__":
    pytest.main()



import matplotlib.pyplot as plt
import numpy as np


def plot_data_and_score(raw_data, score):
    f, ax = plt.subplots(2, 1, figsize=(20, 10))
    ax[0].plot(raw_data)
    ax[0].set_title("time series")
    ax[1].plot(score, "r")
    ax[1].set_title("change score")
    x_grid, y_grid = np.meshgrid(np.arange(len(score)), np.linspace(*ax[0].get_ylim()))
    if np.max(score):
        z_grid = (score/np.max(score))[x_grid]
    else:
        z_grid = score[x_grid]
    ax[0].contourf(x_grid, y_grid, z_grid, alpha=0.5, cmap="BuPu")
    ax[1].set_xlim(ax[0].get_xlim())
    return f



from setuptools import setup
from setuptools import find_packages
from os import path


# read the contents of requirements.txt
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'requirements.txt'), encoding='utf-8') as f:
    requirements = f.read().splitlines()
with open(path.join(this_directory, 'requirements_tests.txt'), encoding='utf-8') as f:
    requirements_tests = f.read().splitlines()

# read content of the readme
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()
    print(long_description)


def main():

    setup(
        name="changepoynt",
        version="0.0.6",
        author="Lucas Weber",
        author_email="weber-lucas@web.de",
        url="https://github.com/Lucew/changepoynt",
        description="Several change point detection methods implemented in python.",
        long_description=long_description,
        long_description_content_type='text/markdown',
        zip_safe=False,
        include_package_data=True,
        packages=find_packages(exclude=['tests']),
        install_requires=requirements,
        dependency_links=[],
        tests_require=requirements_tests,
        extras_require={'test': requirements_tests},
        setup_requires=[],
        license="BSD License",
        test_suite="tests",
        classifiers=[
            "Programming Language :: Python :: 3.9",
            "License :: OSI Approved :: BSD License",
            "Topic :: Scientific/Engineering",
        ],
        keywords='changepoint timeseries engineering'
    )


if __name__ == '__main__':
    main()



"""
Relative Unconstrained Least-Squares Fitting (RuLSIF): A Python Implementation

[1] Liu S , Yamada M , Collier N , et al.
Change-Point Detection in Time-Series Data by Relative Density-Ratio Estimation[J].
2012.

[2] Kawahara Y , Sugiyama M .
Sequential change‐point detection based on direct density‐ratio estimation[M].
John Wiley & Sons, Inc. 2012.

[3] Kawahara Y , Yairi T , Machida K .
Change-Point Detection in Time-Series Data Based on Subspace Identification[C]
Data Mining, 2007. ICDM 2007. Seventh IEEE International Conference on. IEEE, 2007.

Taken from https://github.com/chenxingqiang/rulsif_abrupt-change_detection
copied and heavily modified for performance and jit compilation.
Also corrected a missing transpose in the Gaussian Kernel class.
Additionally, optimized the cross validation to not recompute the gaussian kernel several times
as we only change a factor within the exponent (the sigma).

Has been compared to the demo implementation of the original author:
https://github.com/anewgithubname/change_detection

Unfortunately, the cross validation can't be JIT compiled, as we need to compute theta_hat, which is way faster
using a specialized scipy function for hermitian matrices. Numba can't deal with scipy yet.
"""
import numpy as np
import scipy
import numba as nb
from warnings import warn
from scipy.linalg._misc import LinAlgError, LinAlgWarning
from scipy.linalg.lapack import get_lapack_funcs, _compute_lwork


def _solve_check(info, lamch=None, rcond=None):
    """
    Stolen from scipy to use scipy.linalg.solve
    Check arguments during the different steps of the solution phase
    """
    if info < 0:
        raise ValueError('LAPACK reported an illegal value in {}-th argument'
                         '.'.format(-info))
    elif 0 < info:
        raise LinAlgError('Matrix is singular.')

    if lamch is None:
        return
    E = lamch('E')
    if rcond < E:
        warn('Ill-conditioned matrix (rcond={:.6g}): '
             'result may not be accurate.'.format(rcond),
             LinAlgWarning, stacklevel=3)


def solve(a, b, lower=False, overwrite_a=False, overwrite_b=False):
    """
    DANGER!
    This implementation is stolen from within the scipy package and is equivalent to scipy.linalg.solve
    We reduced a lot of checks to tailor it to the given case within this project and make it as fast as possible
    as it is in the hottest path of RuLSIF and by far the *most called function*.

    This might result in problems, as the checks are really minimal.
    """

    # calculate matrix norm using LAPACK
    lange = get_lapack_funcs('lange', (a,))
    anorm = lange('1', a)

    # this is the code stolen from scipy.linalg.solve for the option assume_a = "sym"
    sycon, sysv, sysv_lw = get_lapack_funcs(('sycon', 'sysv', 'sysv_lwork'), (a, b))
    lu, ipvt, x, info = sysv(a, b, lwork=_compute_lwork(sysv_lw, a.shape[0], lower),
                             lower=lower, overwrite_a=overwrite_a, overwrite_b=overwrite_b)

    # check whether the solver worked in LAPACK
    _solve_check(info)

    # can't tell what this is, but seems like we need it
    rcond, info = sycon(lu, ipvt, anorm)

    # Get the correct lamch function for doubles (can't use any other!)
    lamch = get_lapack_funcs('lamch', dtype='d')
    _solve_check(info, lamch, rcond)
    return x


@nb.njit()
def compute_distance(samples: np.ndarray, sample_means: np.ndarray) -> np.ndarray:
    """
    Compute the distances between points in the sample's feature space
    to points along the center of the distribution
    """

    # get the amount of samples
    (sd, sample_cols) = samples.shape
    (md, mean_cols) = sample_means.shape

    # compute the squared sums
    squared_samples = np.sum(samples ** 2, 0)
    squared_means = np.sum(sample_means ** 2, 0)

    # compute the kernel matrix by using (a-b)^2 = a^2 - 2ab + b^2
    """kernel_matrix = np.tile(squared_means, (sample_cols, 1))\
                    + np.tile(squared_samples, (mean_cols, 1)).T\
                    - 2 * np.dot(samples.T, sample_means)"""
    repeated_squared_samples = np.zeros((mean_cols, sample_cols))
    for idx in range(mean_cols):
        repeated_squared_samples[idx, :] = squared_samples
    repeated_squared_means = np.zeros((sample_cols, mean_cols))
    for idx in range(sample_cols):
        repeated_squared_means[idx, :] = squared_means
    kernel_matrix = repeated_squared_means + repeated_squared_samples.T - 2 * samples.T@sample_means

    return kernel_matrix


@nb.njit()
def compute_gaussian_kernel(samples: np.ndarray, sample_means: np.ndarray, sigma: float) -> np.ndarray:
    """
    Computes an n-dimensional Gaussian/RBF kernel matrix by taking points
    in the sample's feature space and maps them to kernel coordinates in
    Hilbert space by calculating the distance to each point in the sample
    space and taking the Gaussian function of the distances.
       K(X,Y) = exp( -(|| X - Y ||^2) / (2 * sigma^2) )
    where X is the matrix of data points in the sample space,
          Y is the matrix of gaussian centers in the sample space
         sigma is the width of the gaussian function being used
    """
    squared_distance = compute_distance(samples, sample_means)
    # squared_distance = scipy.spatial.distance_matrix(samples.T, sample_means.T)**2
    return np.exp(-squared_distance / (2 * (sigma ** 2)))


@nb.njit()
def update_sigma_gaussian_kernel(old_kernel_values: np.ndarray, old_sigma: float, new_sigma: float):

    # the gaussian kernel ist exp(-dist/(2*sigma^2)), if we want to change the sigma we do not need
    # to recompute since it is only an factor in an exponent.
    #
    # using the exponential rules:
    # (e^a)^b= e^(a*b), we can update the sigma using the following update rule
    # exp(-dist/(2*sigma_new^2)) = exp(-dist/(2*sigma_old^2))*(sigma_old^2/sigma_new^2)
    return old_kernel_values**((old_sigma**2)/(new_sigma**2))


@nb.njit()
def compute_H_hat(alpha=0.0, kernel_matrix_ref_samples=None, kernel_matrix_test_samples=None):
    """
    Calculates the H_hat term of the theta_hat optimization problem
    """
    n_ref = kernel_matrix_ref_samples.shape[1]
    n_test = kernel_matrix_test_samples.shape[1]

    H_hat = (alpha / n_ref) * kernel_matrix_ref_samples@kernel_matrix_ref_samples.T + \
            ((1.0 - alpha) / n_test) * kernel_matrix_test_samples@kernel_matrix_test_samples.T

    return H_hat


@nb.njit()
def compute_h_hat(kernel_matrix_ref_samples):
    """
    Calculates the h_hat term of the theta_hat optimization problem
    """
    h_hat = np.zeros((kernel_matrix_ref_samples.shape[0], 1))
    for idx in range(kernel_matrix_ref_samples.shape[0]):
        h_hat[idx, 0] = kernel_matrix_ref_samples[idx, :].mean()

    # unfortunately this is not numba friendly as axis parameter is not allowed
    # h_hat = np.mean(kernel_matrix_ref_samples, 1, keepdims=True)

    return h_hat


def compute_theta_hat(H_hat: np.ndarray, lambda_scaled_identity, h_hat: np.ndarray):
    """
    Calculates theta_hat given H_hat, h_hat, lambda, and the kernel basis function
    Treat as a system of linear equations and find the exact, optimal
    solution
    """
    # theta_hat = np.linalg.solve(H_hat + (lambda_regularizer * np.eye(kernel_basis)), h_hat)
    # due to the way H_hat is constructed it is symmetric (given some minor numerical errors)
    # see page 11 in [1] for the definition. As it is also real, its is automatically hermitian
    #
    # also we stole and shortened the scipy solve implementation to skip some Checks
    # theta_hat = scipy.linalg.solve(H_hat + (lambda_regularizer * np.eye(kernel_basis)), h_hat, assume_a='her',
    # check_finite=False)
    #
    # creating the identity within this function takes time (as we need to call the numpy function and create
    # the matrix) therefore we do it outside of the function
    return solve(H_hat + lambda_scaled_identity, h_hat)


@nb.njit()
def j_of_theta(alpha: float, g_xref_theta: np.ndarray, g_xtest_theta: np.ndarray):
    """
    Calculates the squared error criterion, J
    """
    return ((alpha / 2.0) * (np.mean(g_xref_theta ** 2)) +
            ((1 - alpha) / 2.0) * (np.mean(g_xtest_theta ** 2)) -
            np.mean(g_xref_theta))


@nb.njit()
def g_of_x_theta(kernel_matrix_samples: np.ndarray, theta_hat: np.ndarray) -> np.ndarray:
    """
    Calculate the alpha-relative density ratio kernel model
    """
    return theta_hat.T@kernel_matrix_samples


class AlphaRelativeDensityRatioEstimator:
    """
    Computes the alpha-relative density ratio estimate of P(X_ref) and P(X_test)
    The alpha-relative density ratio estimator, r_alpha(X), is given by the
    following kernel model:
    g(X; theta) = SUM( (theta_l * K(X, X_centers_l)), l=0, n )
    where theta is a vector of parameters [theta_1, theta_2, ..., theta_l]^T
    to be learned from the data samples. The parameters theta in the model
    g(X; theta) is calculated by solving the following optimization problem:
      theta_hat = argmin [ ( (1/2) * theta^T * H_hat * theta) - (h_hat^T * theta) + ( lambda/2 * theta^T * theta) ]
    where the expression (lambda/2 * theta^T * theta), with lambda >= 0, is
    a regularization term used to penalize against overfitting
    Reference:
    Relative Density-Ratio Estimation for Robust Distribution Comparison. Makoto Yamada,
    Taiji Suzuki, Takafumi Kanamori, Hirotaka Hachiya, and Masashi Sugiyama. NIPS,
    page 594-602. (2011)
    """

    def __init__(self, alpha_constraint=0.0, sigma_width=1.0, lambda_regularizer=0.0, kernel_basis=1):
        self.alpha_constraint = alpha_constraint
        self.sigma_width = sigma_width
        self.lambda_regularizer = lambda_regularizer
        self.kernel_basis = kernel_basis

    def apply(self, reference_samples: np.ndarray, test_samples: np.ndarray, gaussian_centers: np.ndarray) \
            -> (float, float):
        """
        Computes the alpha-relative density ratio, r_alpha(X), of P(X_ref) and P(X_test)
          r_alpha(X) = P(Xref) / (alpha * P(Xref) + (1 - alpha) * P(X_test)
        Returns density ratio estimate at X_ref, r_alpha_ref, and at X_test, r_alpha_test
        """
        # Apply the kernel function to the reference and test samples
        k_ref = compute_gaussian_kernel(reference_samples, gaussian_centers, self.sigma_width).T
        k_test = compute_gaussian_kernel(test_samples, gaussian_centers, self.sigma_width).T

        # Compute the parameters, theta_hat, of the density ratio estimator
        H_hat = compute_H_hat(self.alpha_constraint, k_ref, k_test)
        h_hat = compute_h_hat(k_ref)
        theta_hat = compute_theta_hat(H_hat, self.lambda_regularizer*np.eye(self.kernel_basis), h_hat)

        # Estimate the density ratio, r_alpha_ref = r_alpha(X_ref)
        r_alpha_ref = g_of_x_theta(k_ref, theta_hat)
        # Estimate the density ratio, r_alpha_test = r_alpha(X_test)
        r_alpha_test = g_of_x_theta(k_test, theta_hat)

        return r_alpha_ref, r_alpha_test


class PearsonRelativeDivergenceEstimator:
    """
    Calculates the alpha-relative Pearson divergence score
    The alpha-relative Pearson divergence is given by the following expression:
      PE_alpha = -(alpha/2(n_ref)) * SUM(r_alpha(X_ref_i)^2, i=0, n_ref)        -
                  ((1-alpha)/2(n_test)) * SUM(r_alpha(X_test_j)^2, j=0, n_test) +
                  (1/n_ref) * SUM(r_alpha(X_ref_i), i=0, n_ref)                 -
                  1/2
    where r_alpha(X) is the alpha-relative density ratio estimator and is given by
    the following kernel model:
      g(X; theta) = SUM( (theta_l * K(X, X_centers_l)), l=0, n )
    Reference:
    Relative Density-Ratio Estimation for Robust Distribution Comparison. Makoto
    Yamada, Taiji Suzuki, Takafumi Kanamori, Hirotaka Hachiya, and Masashi Sugiyama.
    NIPS, page 594-602. (2011)
    """

    def __init__(self, alpha_constraint=0.0, sigma_width=1.0, lambda_regularizer=0.0, kernel_basis=1):
        self.alpha_constraint = alpha_constraint
        self.sigma_width = sigma_width
        self.lambda_regularizer = lambda_regularizer
        self.kernel_basis = kernel_basis

    def apply(self, reference_samples=None, test_samples=None, gaussian_centers=None):
        """
        Calculates the alpha-relative Pearson divergence score
        """
        density_ratio_estimator = AlphaRelativeDensityRatioEstimator(self.alpha_constraint, self.sigma_width,
                                                                     self.lambda_regularizer, self.kernel_basis)

        # Estimate alpha relative density ratio and pearson divergence score
        (r_alpha_Xref, r_alpha_Xtest) = density_ratio_estimator.apply(reference_samples, test_samples, gaussian_centers)

        pe_divergence = (np.mean(r_alpha_Xref) -
                         (0.5 * (self.alpha_constraint * np.mean(r_alpha_Xref ** 2) +
                                 (1.0 - self.alpha_constraint) * np.mean(r_alpha_Xtest ** 2))) - 0.5)

        return pe_divergence, r_alpha_Xtest


class RuLSIF:
    """
    Estimates the alpha-relative Pearson Divergence via the least Squares Relative
    Density Ratio Approximation
    Reference:
    Relative Density-Ratio Estimation for Robust Distribution Comparison. Makoto
    Yamada, Taiji Suzuki, Takafumi Kanamori, Hirotaka Hachiya, and Masashi Sugiyama.
    NIPS, page 594-602. (2011)
    """

    def __init__(self, alpha=0.1, kernel_number=100, cross_folds=5, sigma: float = None, lambda_: float = None):
        self.alpha = alpha
        self.kernel_number = kernel_number
        self.cross_folds = cross_folds
        self.gaussian_centers = None
        self.sigma_width = sigma
        self.lambda_regularizer = lambda_

        # check if we need to set sigma or lambda using cross validation
        self.cv = self.sigma_width is None or self.lambda_regularizer is None

    @staticmethod
    def compute_gaussian_width_candidates(reference_samples: np.ndarray, test_samples: np.ndarray):
        """
        Compute a candidate list of Gaussian kernel widths. The best width will be
        selected via cross-validation

        Jaakkola's heuristic method for setting the width parameter of the Gaussian
        radial basis function kernel is to pick a quantile (usually the median) of
        the distribution of Euclidean distances between points having different
        labels.
        Reference:

        Jaakkola T S, Haussler D. Exploiting Generative Models in Discriminative Classifiers[J].
        Advances in Neural Information Processing Systems, 1998, 11(11):487--493.

        Jaakkola, M. Diekhaus, and D. Haussler. Using the Fisher kernel method to detect
        remote protein homologies. In T. Lengauer, R. Schneider, P. Bork, D. Brutlad, J.
        Glasgow, H.- W. Mewes, and R. Zimmer, editors, Proceedings of the Seventh
        International Conference on Intelligent Systems for Molecular Biology.

        It is the same technique as proposed in [1]
        """
        # create a complete sample array
        samples = np.c_[reference_samples, test_samples]

        # compute the pairwise distances
        distances = scipy.spatial.distance.pdist(samples.T, 'sqeuclidean')

        # apply the same formula as in
        # https://github.com/anewgithubname/change_detection/blob/main/lib/comp_med.m
        median_distance = np.sqrt(0.5 * np.median(distances[distances > 0]))

        return median_distance * np.array([0.6, 0.8, 1, 1.2, 1.4])

    @staticmethod
    def generate_regularization_params():
        """
        Generates a candidate list of regularization parameters to be used
        with the L1 regularizer term of RULSIF optimization problem.  The
        best regularizer parameter will be chosen via cross-validation.
        The values itself are taken from the paper [1].
        """
        return 10.0 ** np.array([-3, -2, -1, 0, 1])

    def generate_gaussian_centers(self, reference_samples=None):
        """
        Choose Gaussian centers randomly from the reference samples.
        """
        numcols = reference_samples.shape[1]
        reference_sample_idxs = np.random.permutation(numcols)

        self.kernel_number = min(self.kernel_number, numcols)
        gaussian_centers = reference_samples[:, reference_sample_idxs[0:self.kernel_number]]

        return gaussian_centers

    @staticmethod
    def cross_validate(reference_samples: np.ndarray, test_samples: np.ndarray, gaussian_centers: np.ndarray,
                       sigma_widths: np.ndarray, lambda_candidates: np.ndarray, alpha: float, kernel_number: int,
                       cross_folds=5):
        (refRows, refCols) = reference_samples.shape
        (testRows, testCols) = test_samples.shape

        # Initialize cross validation scoring matrix
        cross_validation_scores = np.zeros((sigma_widths.shape[0], lambda_candidates.shape[0]))

        # Initialize a cross validation index assignment list
        reference_samples_cv_idxs = np.random.permutation(refCols)
        reference_samples_cv_split = (np.arange(start=0, stop=refCols, step=1) * cross_folds) // refCols

        test_samples_cv_idxs = np.random.permutation(testCols)
        test_samples_cv_split = (np.arange(start=0, stop=testCols, step=1) * cross_folds) // testCols

        # initially calculate the kernel matrix using the candidate sigma width
        K_ref = compute_gaussian_kernel(reference_samples, gaussian_centers, sigma_widths[0]).T
        K_test = compute_gaussian_kernel(test_samples, gaussian_centers, sigma_widths[0]).T
        old_sigma = sigma_widths[0]

        # create an identity matrix which we need for computing theta
        # but as we can already create it here, we save time in our hottest path deep in the cross validation
        identity = np.eye(kernel_number)

        # Initiate k-fold cross-validation procedure. Using variable notation similar to the RuLSIF formulas in [1].
        for sigma_idx, sigma in enumerate(sigma_widths):

            # update the kernel for the new sigma without recomputing all distances
            K_ref = update_sigma_gaussian_kernel(K_ref, old_sigma, sigma)
            K_test = update_sigma_gaussian_kernel(K_test, old_sigma, sigma)

            # keep track of the sigma we used, so we keep updating our kernel
            old_sigma = sigma

            for fold_idx in range(cross_folds):

                # get the gaussian kernels for the current fold
                K_ref_trainingSet = K_ref[:, reference_samples_cv_idxs[reference_samples_cv_split != fold_idx]]
                K_test_trainingSet = K_test[:, test_samples_cv_idxs[test_samples_cv_split != fold_idx]]

                # compute the matrices for the current trainings fold
                H_h_KthFold = compute_H_hat(alpha, K_ref_trainingSet, K_test_trainingSet)
                h_h_KthFold = compute_h_hat(K_ref_trainingSet)

                # Select the subset of the kernel matrix not used in the training set
                # for use as the test set to validate against
                K_ref_testSet = K_ref[:, reference_samples_cv_idxs[reference_samples_cv_split == fold_idx]]
                K_test_testSet = K_test[:, test_samples_cv_idxs[test_samples_cv_split == fold_idx]]

                for lambda_idx, lambda_candidate in enumerate(lambda_candidates):
                    # Note: This is the absolute hot path of the complete function, if we can speed up anything
                    # within this loop, we will save a lot of time!

                    # compute the theta for the given parameters
                    theta_h_KthFold = compute_theta_hat(H_h_KthFold, identity*lambda_candidate, h_h_KthFold)

                    # compute the g_of_theta for the given parameters
                    r_alpha_Xref = g_of_x_theta(K_ref_testSet, theta_h_KthFold)
                    r_alpha_Xtest = g_of_x_theta(K_test_testSet, theta_h_KthFold)

                    # Calculate the objective function J(theta) under the current parameters and update the scores
                    cross_validation_scores[sigma_idx, lambda_idx] += j_of_theta(alpha, r_alpha_Xref, r_alpha_Xtest)

        return cross_validation_scores/cross_folds

    def compute_model_parameters(self, reference_samples=None, test_samples=None, gaussian_centers=None):
        """
        Computes model parameters via k-fold cross validation process
        """

        # get the parameter candidates
        sigma_widths = self.compute_gaussian_width_candidates(reference_samples, test_samples)
        lambda_candidates = self.generate_regularization_params()

        # compute the scores
        cross_validation_scores = self.cross_validate(reference_samples, test_samples, gaussian_centers, sigma_widths,
                                                      lambda_candidates, self.alpha, self.kernel_number,
                                                      self.cross_folds)

        # check for the minimal score within the cross_validation
        cv_min_idx_for_sigma, cv_min_idx_for_lambda = np.unravel_index(cross_validation_scores.argmin(),
                                                                       cross_validation_scores.shape)

        # get the optimal values for both parameters
        optimal_sigma = sigma_widths[cv_min_idx_for_sigma]
        optimal_lambda = lambda_candidates[cv_min_idx_for_lambda]
        return optimal_sigma, optimal_lambda

    def train(self, reference_samples=None, test_samples=None):
        """
        Learn the proper model parameters if we did not specify them already
        """

        self.gaussian_centers = self.generate_gaussian_centers(reference_samples)

        # check whether we need to find the model parameters or if they have been given
        if self.cv:
            optimal_sigma, optimal_lambda = self.compute_model_parameters(reference_samples,
                                                                          test_samples, self.gaussian_centers)

            self.sigma_width = optimal_sigma
            self.lambda_regularizer = optimal_lambda

    def apply(self, reference_samples=None, test_samples=None):
        """
        Estimates the alpha-relative Pearson divergence as determined by the relative
        ratio of probability densities:
           P(reference_samples[x]) / (alpha * P(reference_samples[x]) + (1 - alpha) * P(test_samples[x]))
        from samples:
           reference_samples[x_i] | reference_samples[x_i] in R^{d}, with i=1 to reference_samples{N}
        drawn independently of P(reference_samples[x])
        and from samples:
           test_samples[x_j] | test_samples[x_j] in R^{d}, with j=1 to test_samples{N}
        drawn independently from P(test_samples[x])
        After the model hyperparameters have been learned and chosen by the train()
        method, the RULSIF algorithm can be applied repeatedly on both in-sample and out
        of sample data
        """

        if self.gaussian_centers is None or self.kernel_number is None:
            raise Exception("Missing kernel basis function parameters")

        if self.sigma_width == 0.0 or self.lambda_regularizer == 0.0:
            raise Exception("Missing model selection parameters")

        divergence_estimator = PearsonRelativeDivergenceEstimator(self.alpha, self.sigma_width,
                                                                  self.lambda_regularizer, self.kernel_number)
        (pe_alpha, r_alpha_Xtest) = divergence_estimator.apply(reference_samples, test_samples, self.gaussian_centers)

        return pe_alpha

    def __call__(self, reference_samples: np.ndarray, test_samples: np.ndarray):

        # make the normalization as done in
        # https://github.com/anewgithubname/change_detection/blob/main/change_detection.m
        all_samples = np.c_[reference_samples, test_samples]
        std = np.std(all_samples, axis=1) + np.finfo(float).eps
        reference_samples /= std[:, None]
        test_samples /= std[:, None]

        # get the parameters estimation
        self.train(reference_samples, test_samples)

        # compute the result
        return self.apply(reference_samples, test_samples)


def simple_update_test():

    # make a random matrix of signals
    ref = np.random.normal(size=(10, 50))
    test = np.random.normal(size=(10, 50))

    # compute two gaussian kernels
    old_sigma = 1.1
    new_sigma = 3.4
    kernel_old = compute_gaussian_kernel(ref, test, old_sigma)
    kernel_new = compute_gaussian_kernel(ref, test, new_sigma)

    # compare with the kernel update function
    print(np.allclose(kernel_new, update_sigma_gaussian_kernel(kernel_old, old_sigma, new_sigma), atol=1.e-12))


if __name__ == '__main__':
    simple_update_test()



import numpy as np
from changepoynt.algorithms.base_algorithm import Algorithm
import stumpy


class FLUSS(Algorithm):
    """
    This class uses the FLUSS algorithm described in:

    [1]
    Gharghabi, Shaghayegh, et al.
    "Matrix profile VIII: domain agnostic online semantic segmentation at superhuman performance levels."
    2017 IEEE international conference on data mining (ICDM). IEEE, 2017.

    This class essentially wraps the stumpy library which implements a fast approach to calculate the matrix profile.
    https://stumpy.readthedocs.io/en/latest/index.html

    TODO:
    Implement GPU support and streaming? (FLOSS)
    """

    def __init__(self, window_length: int) -> None:
        """
        We initialize the matrix profile and the subsequent floss using only the window length used for comarisons.

        :param window_length: the length of the window used for the distance comparisons in the matrix profile.
        """

        # save the specified parameters into instance variables
        self.window_length = window_length

    def transform(self, time_series: np.ndarray) -> np.ndarray:
        """
        This function calculate the anomaly score for each sample within the time series.

        It also does some assertion regarding time series length.

        :param time_series: 1D array containing the time series to be scored
        :return: anomaly score
        """

        # check the dimensions of the input array
        assert time_series.ndim == 1, "Time series needs to be an 1D array."

        # check that we have at least two windows
        assert time_series.shape[0] > self.window_length, 'Time series needs to be longer than window length.'

        # compute the floss score which is the amount of index pointers over this sample
        # compared to an expected curve. It is capped by one.
        matrix_profile = stumpy.stump(time_series, m=self.window_length)
        corrected_arc_crossings, _ = stumpy.fluss(matrix_profile[:, 1], L=self.window_length, n_regimes=1)
        return 1-corrected_arc_crossings


def _main():
    """
    Internal quick testing function.

    :return:
    """
    from time import time
    # make synthetic step function
    np.random.seed(123)
    # synthetic (frequency change)
    x0 = np.sin(2 * np.pi * 1 * np.linspace(0, 10, 1000))
    x1 = np.sin(2 * np.pi * 2 * np.linspace(0, 10, 1000))
    x2 = np.sin(2 * np.pi * 8 * np.linspace(0, 10, 1000))
    x3 = np.sin(2 * np.pi * 4 * np.linspace(0, 10, 1000))
    x = np.hstack([x0, x1, x2, x3])
    x += np.random.rand(x.size)

    # create the method
    fluss_recognizer = FLUSS(50)

    # compute the score
    start = time()
    score = fluss_recognizer.transform(x)
    print(f'Computation for {len(x)} signal values took {time()-start} s.')


if __name__ == '__main__':
    _main()
